from typing import Sequence, Optional
from datetime import datetime, time

from .utils import (
    parse_mdc_time, pack_mdc_time, parse_enum_bitmask, pack_bitmask,
    pack_videowall_model, parse_videowall_model)


class Field:
    def __init__(self, name=None):
        self.name = name or self.__class__.__name__.upper()

    def parse(self, data):
        # returns data and cursor shift
        return data, len(data)

    def pack(self, value):
        return [value]


class Int(Field):
    range: Optional[range] = None

    def __init__(self, name=None, range=None, length=1, byteorder='big'):
        super().__init__(name)
        self.range = range
        self.length = length
        self.byteorder = byteorder

    def pack(self, value):
        if self.range and value not in self.range:
            raise ValueError('Field not in range', self.name, self.range)
        return int(value).to_bytes(self.length, byteorder=self.byteorder)

    def parse(self, data):
        return int.from_bytes(data[:self.length], self.byteorder), self.length


class Bool(Int):
    range = range(2)

    def parse(self, data):
        return bool(data[0]), 1


class Enum(Field):
    def __init__(self, enum, name=None):
        self.enum = enum
        super().__init__(name or enum.__name__)

    def parse(self, data):
        return self.enum(data[0]), 1

    def pack(self, value):
        if isinstance(value, str):
            value = self.enum[value]
        return [self.enum(value).value]


class Str(Field):
    def __init__(self, name=None, length=None):
        self.length = length
        super().__init__(name)

    def parse(self, data):
        length = self.length or len(data)
        return data[:length].decode('utf8').rstrip('\x00'), length

    def pack(self, value):
        if self.length is not None and len(value) > self.length:
            raise ValueError('Field length exceeded', len(value), self.length)
        return value.encode('utf8')


class StrCoded(Field):
    def __init__(self, code, name=None):
        super().__init__(name)
        self.code = code

    def parse(self, data):
        if self.code != data[0]:
            raise ValueError('Expected code not matched', data[0])
        length = data[1]
        if not length:
            return '', 2
        return data[2:length + 2].decode('utf8'), length + 2

    def pack(self, value):
        encoded = value.encode('utf8')
        return bytes([self.code, len(encoded)]) + encoded


class Time12H(Field):
    def parse(self, data):
        return parse_mdc_time(data[2], data[0], data[1]), 3

    def pack(self, data):
        day_part, hour, minute, second = pack_mdc_time(data)
        return (hour, minute, day_part)


class Time(Field):
    def __init__(self, name=None, seconds=False):
        self.seconds = seconds
        super().__init__(name)

    def parse(self, data):
        return (
            time(data[0], data[1], data[3] if self.seconds else 0),
            3 if self.seconds else 2
        )

    def pack(self, data):
        if self.seconds:
            return (data.hour, data.minute, data.second)
        return (data.hour, data.minute)


class DateTime(Field):
    name = 'datetime'

    def __init__(self, name=None, seconds=True):
        self.seconds = seconds
        super().__init__(name)

    def parse(self, data):
        if self.seconds:
            time = parse_mdc_time(data[7], data[1], data[2], data[3])
            return datetime(
                int.from_bytes(data[5:7], 'big'),  # year
                data[4], data[0],  # month, day
                time.hour, time.minute, time.second
            ), 8

        time = parse_mdc_time(data[6], data[1], data[2])
        return datetime(
                int.from_bytes(data[4:6], 'big'),  # year
                data[3], data[0],  # month, day
                time.hour, time.minute, time.second
        ), 7

    def pack(self, value):
        day_part, hour, minute, second = pack_mdc_time(value.time())
        return (
            bytes([value.day, hour, minute])
            + (self.seconds and bytes([second]) or b'')
            + bytes([value.month])
            + int.to_bytes(value.year, 2, 'big') + bytes([day_part]))


class Bitmask(Enum):
    def parse(self, data):
        return parse_enum_bitmask(self.enum, data[0]), 1

    def pack(self, data):
        if not isinstance(data, Sequence):
            raise ValueError('Bitmask values must be sequence')
        return [pack_bitmask(data)]


class IPAddress(Field):
    def parse(self, data):
        return '.'.join(str(int(x)) for x in data[:4]), 4

    def pack(self, data):
        rv = tuple(int(x) for x in data.split('.'))
        if not len(rv) == 4 or not all(0 <= x < 256 for x in rv):
            raise ValueError('Invalid IP address', data)
        return rv


class VideoWallModel(Field):
    def parse(self, data):
        return parse_videowall_model(data[0]), 1

    def pack(self, data):
        if isinstance(data, str):
            rv = tuple(int(x) for x in data.split(','))
        elif not isinstance(data, (tuple, list)):
            raise TypeError('Video wall model must be Tuple[int, int] or'
                            'comma-separated string')
        if not len(rv) == 2:
            raise ValueError('Invalid video wall model', data)
        return pack_videowall_model(rv)



from typing import Union, Sequence, Tuple
from functools import partial
from enum import Enum
import asyncio

from .exceptions import MDCResponseError, MDCReadTimeoutError, \
    MDCTimeoutError, MDCTLSRequired, MDCTLSAuthFailed
from .utils import repr_hex


HEADER_CODE = 0xAA
RESPONSE_CMD = 0xFF
ACK_CODE = ord('A')  # 0x41 65
NAK_CODE = ord('N')  # 0x4E 78


def get_checksum(payload):
    # payload should be without HEADER_CODE
    return sum(payload) % 256


def _normalize_cmd(
    cmd: Union[int, Tuple[int], Tuple[int, Union[int, None]]]
) -> Tuple[int, Union[int, None]]:
    """
    Returns (cmd, subcmd) tuple
    """
    if isinstance(cmd, int):
        return cmd, None
    elif len(cmd) == 1:
        return int(cmd[0]), None
    elif not len(cmd) == 2:
        raise ValueError('cmd tuple should contain only (cmd, subcmd)')
    return int(cmd[0]), None if cmd[1] is None else int(cmd[1])


def pack_payload(
    cmd: Union[int, Tuple[int], Tuple[int, Union[int, None]]],
    display_id: int,
    data: Union[bytes, Sequence] = b''
):
    cmd, subcmd = _normalize_cmd(cmd)
    data = bytes(data)
    if subcmd is not None:
        data = bytes([subcmd]) + data

    payload = (
        bytes([HEADER_CODE, cmd, display_id, len(data)])
        + bytes(data)
    )
    payload += bytes([get_checksum(payload[1:])])
    return payload


def pack_response(
    cmd: Union[int, Tuple[int], Tuple[int, int]],
    display_id: int,
    ack: bool,
    data: Union[bytes, Sequence] = b''
):
    cmd, subcmd = _normalize_cmd(cmd)

    if not ack and len(data) != 1:
        raise ValueError(
            'Data should contain only error code for NAK response', data)

    data = bytes(data)
    if subcmd is not None and ack:
        # subcmd is not sent on NAK in response
        data = bytes([subcmd]) + data

    return pack_payload(
       RESPONSE_CMD, display_id,
       bytes([ACK_CODE if ack else NAK_CODE, cmd]) + data
    )


async def wait_for(aw, timeout, reason):
    try:
        return await asyncio.wait_for(aw, timeout)
    except asyncio.TimeoutError as exc:
        raise MDCTimeoutError(reason) from exc


async def wait_for_read(reader, count, timeout, reason):
    try:
        return await asyncio.wait_for(reader.read(count), timeout)
    except asyncio.TimeoutError as exc:
        raise MDCReadTimeoutError(reason, bytes(reader._buffer)) from exc


class CONNECTION_MODE(Enum):
    TCP = 'tcp'
    SERIAL = 'serial'


class MDCConnection:
    reader, writer = None, None

    def __init__(self, target, mode=CONNECTION_MODE.TCP, timeout=5,
                 connect_timeout=None, verbose=False, **connection_kwargs):
        self.target = target
        self.mode = CONNECTION_MODE(mode)
        self.connection_kwargs = connection_kwargs

        self.timeout = timeout
        self.connect_timeout = connect_timeout or timeout
        self.verbose = (
            partial(print, self.target) if verbose is True else verbose)

    async def open(self):
        connection_kwargs = self.connection_kwargs.copy()
        pin = connection_kwargs.pop('pin', None)

        if self.mode == CONNECTION_MODE.TCP:
            if isinstance(self.target, (list, tuple)):
                # make target be compatible with socket.__init__
                target, port = self.target
            else:
                target, *port = self.target.split(':')
                port = port and int(port[0]) or 1515
            connection_kwargs.setdefault('port', port)

            self.reader, self.writer = \
                await wait_for(
                    asyncio.open_connection(target, **connection_kwargs),
                    self.connect_timeout, 'Connect timeout')

            if self.verbose:
                self.verbose('Connected')

        else:
            # Make this package optional
            from serial_asyncio import (  # type: ignore[import-untyped]
                open_serial_connection
            )

            self.reader, self.writer = \
                await wait_for(
                    open_serial_connection(
                        url=self.target,
                        **connection_kwargs),
                    self.connect_timeout, 'Connect timeout')

            if self.verbose:
                self.verbose('Connected')

        if pin is not None:
            try:
                await self._start_tls(pin)
            except Exception:
                await self.close()
                raise

    async def _start_tls(self, pin):
        if isinstance(pin, int):
            pin = str(pin).rjust(4, '0').encode()
        elif isinstance(pin, str):
            pin = pin.rjust(4, '0').encode()

        assert self.is_opened

        import ssl
        resp = await wait_for_read(self.reader, 15, self.timeout,
                                   'TLS header read timeout')
        if not resp == b'MDCSTART<<TLS>>':
            raise MDCResponseError('Unexpected TLS header',
                                   resp + self.reader._buffer)
        transport = self.writer.transport
        protocol = transport.get_protocol()
        loop = asyncio.get_event_loop()
        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ssl_ctx.minimum_version = ssl.TLSVersion.MINIMUM_SUPPORTED
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.VerifyMode.CERT_NONE
        ssl_transport = await loop.start_tls(
            transport, protocol, ssl_ctx)
        self.writer._transport = ssl_transport
        self.reader._transport = ssl_transport

        if self.verbose:
            self.verbose('TLS established')

        self.writer.write(pin)
        await wait_for(self.writer.drain(), self.timeout,
                       'Write pin timeout')
        resp = await wait_for_read(self.reader, 15, self.timeout,
                                   'TLS auth read timeout')
        if not resp == b'MDCAUTH<<PASS>>':
            if resp[:14] == b'MDCAUTH<<FAIL:':
                resp += await wait_for_read(self.reader, 5, self.timeout,
                                            'TLS auth fail read timeout')
                if resp[-2:] != b'>>':
                    raise MDCResponseError('Unexpected TLS auth fail response',
                                           resp + self.reader._buffer)
                try:
                    fail_code = int(resp[-6:-2], 16)
                except ValueError:
                    raise MDCResponseError('Unexpected TLS auth fail code',
                                           resp + self.reader._buffer)
                raise MDCTLSAuthFailed(fail_code)
            raise MDCResponseError('Unexpected TLS auth response',
                                   resp + self.reader._buffer)

        if self.verbose:
            self.verbose('TLS authentication passed')

    @property
    def is_opened(self):
        return self.writer is not None

    @property
    def is_tls_started(self):
        return (self.writer and self.writer.transport.__class__.__name__ ==
                '_SSLProtocolTransport')

    async def send(
        self,
        cmd: Union[int, Tuple[int], Tuple[int, int]],
        display_id: int,
        data: Union[bytes, Sequence] = b''
    ):
        cmd, subcmd = _normalize_cmd(cmd)
        payload = pack_payload((cmd, subcmd), display_id, data)

        if not self.is_opened:
            await self.open()
        assert (self.reader is not None and self.writer is not None)

        self.writer.write(payload)
        await wait_for(self.writer.drain(), self.timeout, 'Write timeout')
        if self.verbose:
            self.verbose('Sent', repr_hex(payload))

        resp = await wait_for_read(self.reader, 4, self.timeout,
                                   'Response header read timeout')
        if not resp:
            raise MDCResponseError('Empty response', resp)
        if resp[0] != HEADER_CODE:
            if (resp + self.reader._buffer) == b'MDCSTART<<TLS>>':
                raise MDCTLSRequired(resp + self.reader._buffer)
            raise MDCResponseError('Unexpected header',
                                   resp + self.reader._buffer)
        if resp[1] != RESPONSE_CMD:
            raise MDCResponseError('Unexpected cmd',
                                   resp + self.reader._buffer)
        if resp[2] != display_id:
            raise MDCResponseError('Unexpected display_id',
                                   resp + self.reader._buffer)

        length = resp[3]
        resp += await wait_for_read(self.reader, length + 1, self.timeout,
                                    'Response data read timeout')
        if self.verbose:
            self.verbose('Recv', repr_hex(resp))

        checksum = get_checksum(resp[1:-1])
        if checksum != int(resp[-1]):
            raise MDCResponseError('Checksum failed', resp)

        ack, rcmd, data = resp[4], resp[5], resp[6:-1]
        if ack not in (ACK_CODE, NAK_CODE):
            raise MDCResponseError('Unexpected ACK/NAK', resp)

        if subcmd and ack == ACK_CODE:
            # rsubcmd is not sent on NAK
            rsubcmd = data[0]
            data = data[1:]
        else:
            rsubcmd = None

        return (
            ack == ACK_CODE,
            (rcmd,) if rsubcmd is None else (rcmd, rsubcmd),
            data
        )

    async def close(self):
        if self.is_tls_started:
            # FIX warning
            # "returning true from eof_received() has no effect when using ssl"
            self.writer._protocol.eof_received = lambda: None
        writer = self.writer
        self.reader, self.writer = None, None
        writer.close()
        await wait_for(writer.wait_closed(), self.timeout, 'Close timeout')

    async def __aenter__(self):
        if not self.is_opened:
            await self.open()
        return self

    async def __aexit__(self, *args):
        if self.is_opened:
            await self.close()

    def __await__(self):
        return self.__aenter__().__await__()
